__author__ = 'Lukas Leufen'
__date__ = '2020-11-05'

from mlair.data_handler.data_handler_single_station import DataHandlerSingleStation
from mlair.data_handler.data_handler_kz_filter import DataHandlerKzFilterSingleStation
from mlair.data_handler import DefaultDataHandler
from mlair import helpers
from mlair.helpers import remove_items
from mlair.configuration.defaults import DEFAULT_SAMPLING

import inspect
from typing import Callable
import datetime as dt

import numpy as np
import pandas as pd
import xarray as xr


class DataHandlerMixedSamplingSingleStation(DataHandlerSingleStation):
    _requirements = remove_items(inspect.getfullargspec(DataHandlerSingleStation).args, ["self", "station"])

    def __init__(self, *args, sampling_inputs, **kwargs):
        sampling = (sampling_inputs, kwargs.get("sampling", DEFAULT_SAMPLING))
        kwargs.update({"sampling": sampling})
        super().__init__(*args, **kwargs)

    def setup_samples(self):
        """
        Setup samples. This method prepares and creates samples X, and labels Y.
        """
        self._data = list(map(self.load_and_interpolate, [0, 1]))  # load input (0) and target (1) data
        self.set_inputs_and_targets()
        if self.do_transformation is True:
            self.call_transform()
        self.make_samples()

    def load_and_interpolate(self, ind) -> [xr.DataArray, pd.DataFrame]:
        data, self.meta = self.load_data(self.path[ind], self.station, self.statistics_per_var, self.sampling[ind],
                                         self.station_type, self.network, self.store_data_locally, self.data_origin,
                                         self.start, self.end)
        data = self.interpolate(data, dim=self.time_dim, method=self.interpolation_method,
                                limit=self.interpolation_limit)
        return data

    def set_inputs_and_targets(self):
        inputs = self._data[0].sel({self.target_dim: helpers.to_list(self.variables)})
        targets = self._data[1].sel({self.target_dim: self.target_var})
        self.input_data.data = inputs
        self.target_data.data = targets

    def setup_data_path(self, data_path, sampling):
        """Sets two paths instead of single path. Expects sampling arg to be a list with two entries"""
        assert len(sampling) == 2
        return list(map(lambda x: super(__class__, self).setup_data_path(data_path, x), sampling))


class DataHandlerMixedSampling(DefaultDataHandler):
    """Data handler using mixed sampling for input and target."""

    data_handler = DataHandlerMixedSamplingSingleStation
    data_handler_transformation = DataHandlerMixedSamplingSingleStation
    _requirements = data_handler.requirements()


class DataHandlerMixedSamplingWithFilterSingleStation(DataHandlerMixedSamplingSingleStation,
                                                      DataHandlerKzFilterSingleStation):
    _requirements1 = DataHandlerKzFilterSingleStation.requirements()
    _requirements2 = DataHandlerMixedSamplingSingleStation.requirements()
    _requirements = list(set(_requirements1 + _requirements2))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _check_sampling(self, **kwargs):
        assert kwargs.get("sampling") == ("hourly", "daily")

    def setup_samples(self):
        """
        Setup samples. This method prepares and creates samples X, and labels Y.

        A KZ filter is applied on the input data that has hourly resolution. Lables Y are provided as aggregated values
        with daily resolution.
        """
        self._data = list(map(self.load_and_interpolate, [0, 1]))  # load input (0) and target (1) data
        self.set_inputs_and_targets()
        self.apply_kz_filter()
        if self.do_transformation is True:
            self.call_transform()
        self.make_samples()

    def estimate_filter_width(self):
        """
        f = 0.5 / (len * sqrt(itr)) -> T = 1 / f
        :return:
        """
        return int(self.kz_filter_length[0] * np.sqrt(self.kz_filter_iter[0]) * 2)

    @staticmethod
    def _add_time_delta(date, delta):
        new_date = dt.datetime.strptime(date, "%Y-%m-%d") + dt.timedelta(hours=delta)
        return new_date.strftime("%Y-%m-%d")

    def load_and_interpolate(self, ind) -> [xr.DataArray, pd.DataFrame]:

        if ind == 0:  # for inputs
            estimated_filter_width = self.estimate_filter_width()
            start = self._add_time_delta(self.start, -estimated_filter_width)
            end = self._add_time_delta(self.end, estimated_filter_width)
        else:  # target
            start, end = self.start, self.end

        data, self.meta = self.load_data(self.path[ind], self.station, self.statistics_per_var, self.sampling[ind],
                                         self.station_type, self.network, self.store_data_locally, self.data_origin,
                                         start, end)
        data = self.interpolate(data, dim=self.time_dim, method=self.interpolation_method,
                                limit=self.interpolation_limit)
        return data


class DataHandlerMixedSamplingWithFilter(DefaultDataHandler):
    """Data handler using mixed sampling for input and target. Inputs are temporal filtered."""

    data_handler = DataHandlerMixedSamplingWithFilterSingleStation
    data_handler_transformation = DataHandlerMixedSamplingWithFilterSingleStation
    _requirements = data_handler.requirements()


class DataHandlerSeparationOfScalesSingleStation(DataHandlerMixedSamplingWithFilterSingleStation):
    """
    Data handler using mixed sampling for input and target. Inputs are temporal filtered and depending on the
    separation frequency of a filtered time series the time step delta for input data is adjusted (see image below).

    .. image:: ../../../../../_source/_plots/separation_of_scales.png
        :width: 400

    """

    _requirements = DataHandlerMixedSamplingWithFilterSingleStation.requirements()

    def __init__(self, *args, time_delta=np.sqrt, **kwargs):
        assert isinstance(time_delta, Callable)
        self.time_delta = time_delta
        super().__init__(*args, **kwargs)

    def make_history_window(self, dim_name_of_inputs: str, window: int, dim_name_of_shift: str) -> None:
        """
        Create a xr.DataArray containing history data.

        Shift the data window+1 times and return a xarray which has a new dimension 'window' containing the shifted
        data. This is used to represent history in the data. Results are stored in history attribute.

        :param dim_name_of_inputs: Name of dimension which contains the input variables
        :param window: number of time steps to look back in history
                Note: window will be treated as negative value. This should be in agreement with looking back on
                a time line. Nonetheless positive values are allowed but they are converted to its negative
                expression
        :param dim_name_of_shift: Dimension along shift will be applied
        """
        window = -abs(window)
        data = self.input_data.data
        self.history = self.stride(data, dim_name_of_shift, window)

    def stride(self, data: xr.DataArray, dim: str, window: int) -> xr.DataArray:

        # this is just a code snippet to check the results of the kz filter
        # import matplotlib
        # matplotlib.use("TkAgg")
        # import matplotlib.pyplot as plt
        # xr.concat(res, dim="filter").sel({"variables":"temp", "Stations":"DEBW107", "datetime":"2010-01-01T00:00:00"}).plot.line(hue="filter")

        time_deltas = np.round(self.time_delta(self.cutoff_period)).astype(int)
        start, end = window, 1
        res = []
        window_array = self.create_index_array('window', range(start, end), squeeze_dim=self.target_dim)
        for delta, filter_name in zip(np.append(time_deltas, 1), data.coords["filter"]):
            res_filter = []
            data_filter = data.sel({"filter": filter_name})
            for w in range(start, end):
                res_filter.append(data_filter.shift({dim: -w * delta}))
            res_filter = xr.concat(res_filter, dim=window_array).chunk()
            res.append(res_filter)
        res = xr.concat(res, dim="filter")
        return res

    def estimate_filter_width(self):
        """
        Attention: this method returns the maximum value of
        * either estimated filter width f = 0.5 / (len * sqrt(itr)) -> T = 1 / f or
        * time delta method applied on the estimated filter width mupliplied by window_history_size
        to provide a sufficiently wide filter width.
        """
        est = self.kz_filter_length[0] * np.sqrt(self.kz_filter_iter[0]) * 2
        return int(max([self.time_delta(est) * self.window_history_size, est]))


class DataHandlerSeparationOfScales(DefaultDataHandler):
    """Data handler using mixed sampling for input and target. Inputs are temporal filtered and different time step
    sizes are applied in relation to frequencies."""

    data_handler = DataHandlerSeparationOfScalesSingleStation
    data_handler_transformation = DataHandlerSeparationOfScalesSingleStation
    _requirements = data_handler.requirements()
