"""Helper functions that are used to simplify testing."""
import re
from typing import Union, Pattern, List

import numpy as np
import xarray as xr


class PyTestRegex:
    r"""
    Assert that a given string meets some expectations.

    Use like

        >>> PyTestRegex(r"TestString\d+") == "TestString"
        False
        >>> PyTestRegex(r"TestString\d+") == "TestString2"
        True


    :param pattern: pattern or string to use for regular expresssion
    :param flags: python re flags
    """

    def __init__(self, pattern: Union[str, Pattern], flags: int = 0):
        """Construct PyTestRegex."""
        self._regex = re.compile(pattern, flags)

    def __eq__(self, actual: str) -> bool:
        """Return whether regex matches given string actual or not."""
        return bool(self._regex.match(actual))

    def __repr__(self) -> str:
        """Show regex pattern."""
        return self._regex.pattern


class PyTestAllEqual:
    """
    Check if all elements in list are the same.

    :param check_list: list with elements to check
    """

    def __init__(self, check_list: List):
        """Construct class."""
        self._list = check_list
        self._test_function = None

    def _set_test_function(self):
        if isinstance(self._list[0], np.ndarray):
            self._test_function = np.testing.assert_array_equal
        else:
            self._test_function = xr.testing.assert_equal

    def _check_all_equal(self) -> bool:
        """
        Check if all elements are equal.

        :return boolean if elements are equal
        """
        equal = True
        self._set_test_function()
        for b in self._list:
            equal *= self._test_function(self._list[0], b) is None
        return bool(equal == 1)

    def is_true(self) -> bool:
        """
        Start equality check.

        :return: true if equality test is passed, false otherwise
        """
        return self._check_all_equal()


def xr_all_equal(check_list: List) -> bool:
    """
    Check if all given elements (preferably xarray's) in list are equal.

    :param check_list: list with elements to check

    :return: boolean if all elements are the same or not
    """
    equal = True
    for b in check_list:
        equal *= xr.testing.assert_equal(check_list[0], b) is None
    return equal == 1